functor Globals(
	    structure Wr: WR
	    structure Pp: PP
	    structure WrMgt: WRMGT
	    structure LatexWr: LATEXWR
	    structure ListUtils: LISTUTILS
	    structure StringUtils: STRINGUTILS
	    structure DebugUtils: DEBUGUTILS
	    structure Id: ID
	    structure Registry: REGISTRY
	    sharing Pp.Wr = Wr
	        and WrMgt.Pp = Pp
	    ) : GLOBALS 
	    = struct

structure Wr = Wr;
open Wr;

structure Pp = Pp;
open Pp;

structure WrMgt = WrMgt;
open WrMgt;

structure Id = Id;

structure Registry = Registry

open ListUtils
open StringUtils
open DebugUtils

exception CantHappen
exception NotImplemented of string

(* ---------------------------------------------------------------------- *)

val cmds = ref (Registry.new: (string option) Registry.registry)

fun register_cmd name proc =
  cmds := Registry.register (!cmds) name proc

fun do_cmd n a = Registry.poke_one (!cmds) n a

(* ---------------------------------------------------------------------- *)

val resetprocs = ref (nil : ((unit->unit) list))

fun resetproc p = resetprocs := p::(!resetprocs)

fun global_reset () = let
  fun r [] = ()
    | r (p::tl) = (p(); r tl)
  in r (!resetprocs)
  end

(* ---------------------------------------------------------------------- *)

val finalizeprocs = ref (nil : ((unit->unit) list))

fun finalizeproc p = finalizeprocs := p::(!finalizeprocs)

fun global_finalize () = let
  fun r [] = ()
    | r (p::tl) = (p(); r tl)
  in r (!finalizeprocs)
  end

(* ---------------------------------------------------------------------- *)

val flags = ref (Registry.new: bool Registry.registry)

fun set_flag n b = Registry.poke_one (!flags) n b

fun registerflag name flagref =
  flags := Registry.register (!flags) name (fn b => flagref := b)

fun createflag names init = let
  val f = ref init
  in map (fn n => registerflag n f) names;
     resetproc (fn()=> f := init);
     f
  end

(* ---------------------------------------------------------------------- *)

val statsprocs = ref (Registry.new: unit Registry.registry)

fun registerstatsproc name p =
  statsprocs := Registry.register (!statsprocs) name p

fun statsproc names p = 
  (map (fn n => registerstatsproc n p) names;
   ())

fun stats name = Registry.poke_one (!statsprocs) name ()

fun allstats () = Registry.poke_all (!statsprocs) ()


(* ---------------------------------------------------------------------- *)
(* Register some flags and reset procs for lower-level modules that didn't 
   know how to do it themselves *)

val _ = registerflag "latex" LatexWr.LATEX_MODE
val _ = registerflag "pp" Pp.DEBUG

val _ = resetproc reset_wrmgt
val _ = resetproc reset_debugutils
val _ = resetproc (fn()=> Pp.DEBUG := false);

end

