(* How is this guy going to name the begintext and endtext functions? *)

functor Parse (structure Globals : GLOBALS
	       structure ParseRes : PARSERESCOMMON
	       structure Interface : INTERFACE
	       structure Parser : PARSER
	          sharing type Parser.pos = Interface.pos
		  sharing type Parser.result = ParseRes.T
		  sharing type Parser.arg = unit
	       structure Tokens : MY_TOKENS
	          sharing type Tokens.token = Parser.Token.token
		  sharing type Tokens.svalue = Parser.svalue
               ) : PARSE =
struct

structure ParseRes = ParseRes
open Globals

val begin_parser_context = Interface.begin_parser_context
val end_parser_context = Interface.end_parser_context
val ParseError = Parser.ParseError

(* BCP 13.Jun.92 -- In the original version (by D. Tarditi) of this
   function, it used to compare the next token on the lexer returned
   by the parser against Tokens.EOF and call the parser again if 
   they didn't match.  Since I don't see why this test should ever
   be false, I'm removing it. *)
fun parse (lookahead, reader:int->string) = let
  fun invoke lexer = Parser.parse(lookahead,lexer,Interface.error,())
  val (result,lexer) = invoke (Parser.makeLexer reader)
  in result
  end

fun string_reader s =
 let val next = ref s
 in fn _ => !next before next := ""
 end
    
val string_parse = fn s => parse (0, string_reader s)

fun file_parse name = let 
  val dev = open_in name
  in (parse (15,(fn i => input(dev,i)))) 
     before close_in dev
  end

fun cnvt_internal_progquotes line = 
  replace_string "<<" "\\verb$" 
    (replace_string ">>" "$" line)

(* Important: this has the same name as the flag in LatexWr.sml *)
val LATEX_MODE = createflag ["latex"] false

val echo_input = createflag ["echo"] false
fun echo s = if !echo_input then write s else ()

val prompt_char = ref(">");
val _ = register_cmd "prompt" (fn SOME(s) => prompt_char := s | _ => ());

fun echo_line line =
  if !LATEX_MODE then (
  (* Echo and strip for later formatting by latex *)
    if (line = "\n") orelse (line="")
       then (begintext(); write line; endtext(); line)
    else if prefix line 3 = "%%%"
       then (substring(line,3,size(line)-3))
    else if prefix line 2 = (!prompt_char ^ ">")
       then (substring(line,2,size(line)-2))
    else if prefix line 2 = (!prompt_char ^ ">")
       then (begintext(); write (substring(line,2,size(line)-2)); 
             endtext(); "")
    else if prefix line 1 = (!prompt_char)
       then (write line; (substring(line,1,size(line)-1)))
    else (begintext(); write (cnvt_internal_progquotes line); endtext(); "")
  ) else (
  (* Echo and strip as plain text *)
    if (line = "\n") orelse (line="")
       then (echo line; line)
    else if prefix line 3 = "%%%"
       then (substring(line,3,size(line)-3))
    else if prefix line 3 = "%% "
       then (echo (substring(line,3,size(line)-3)); line)
    else if prefix line 2 = "%%"
       then (echo (substring(line,2,size(line)-2)); line)
    else (echo ((!prompt_char ^ " ") ^ line); line)
  )

fun convert_tabs s =
    implode (map (fn "\t" => "        " | s => s) (explode s));

fun stream_parse dev =
   parse (15,(fn i => 
   		 let val line = convert_tabs(input_line(dev))
   		     val stripped = echo_line line
   		 in stripped
   		 end))

fun top_parse () = let
  (* First prompt is supplied by the "done" function in Main.sml *)
  val first = ref(true)
  in parse (0,
       fn i => 
        (if (!first)
            then (first := false)
            else (output (std_out, "- "); flush_out std_out);
         input_line std_in))
  end

end

