signature WRMGT = sig

(* Maintains a notion of a current prettyprinting stream and and its 
   associated writer, and provides operations for performing output
   and formatting operations on them. *)

structure Wr: WR
structure Pp: PP

sharing Pp.Wr = Wr

(* ---------------------------------------------------------------------- *)
(* Management *)

(* Establish a new standard writer.  The given writer is used as the
   target for a new prettyprinter, which is coerced (using Pp.wr_from_pp)
   again into a writer.  This, in turn, becomes the basis for a new latex
   writer, which is again coerced into a writer.  The final Wr.T and 
   its corresponding LatexWr.T and underlying Pp become the new standard 
   streams.  (This would all be much simpler if SML provided subtyping.
   Think of building a new prettyprinter on top of the given writer
   and then building a latex writer on the pretty printer.  The other
   two writers are just simulating subtyping.) *) 
val set_stdwr: Wr.Wr -> unit

(* $$$ Nuke these from the signature 
(* The current standard prettyprinter *)
val stdpp: unit -> Pp.Pp

(* The current standard prettyprinter "viewed as" a writer *)
val stdwr: unit -> Wr.Wr
*)

(* Like set_stdwr, but push the current streams onto a stack *)
val push_stdwr: Wr.Wr -> unit

(* Close the current streams and restores the most-recently saved ones *)
val pop_stdwr: unit -> unit


(* Close all streams *)
val close_wrs: unit -> unit


(* ---------------------------------------------------------------------- *)
(* Utilities *)

(* Wr functions *)
val write: string -> unit
val flush: unit -> unit

(* Pp functions *)
val break: bool -> int -> unit
val ubreak: int -> unit
val uubreak: int -> unit
val setb: unit -> unit
val endb: unit -> unit
val expbreak: bool -> string -> unit
val setmargin: int -> unit

(* LatexWr functions *)
val begintext: unit -> unit
val endtext: unit -> unit

(* ---------------------------------------------------------------------- *)

(* Internal reset routine, called only from Globals *)
val reset_wrmgt: unit -> unit

end
