structure Tokens = Tokens
structure Interface = Interface
open Interface

type pos = Interface.pos
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult= (svalue,pos) token

val eof = fn () => Tokens.T_EOF(!line,!line)

val str_begin = ref(!line);
val str_const = ref([]:string list);

%%
%header (functor MYLexFun(structure Tokens: MY_TOKENS structure Interface: INTERFACE) : LEXER);

%s COMMENT STRING;
idinit=[A-Za-z0-9#^];
idchars=[A-Za-z0-9_'!#^];
digit=[0-9];
id={idinit}{idchars}*;
int={digit}{digit}*;
ws=[\t\ ]*;
string1="\"".*"\"";
string2="`".*"`";
%%
<INITIAL>{ws}	=> (lex());
<INITIAL>\n	=> (next_line(); lex());
<INITIAL>"%"	=> (YYBEGIN COMMENT; lex());
<INITIAL>":"	=> (Tokens.T_COLON(!line,!line));
<INITIAL>"::"	=> (Tokens.T_COLONCOLON(!line,!line));
<INITIAL>":="	=> (Tokens.T_COLONEQ(!line,!line));
<INITIAL>"$"	=> (Tokens.T_DOLLAR(!line,!line));
<INITIAL>"*"	=> (Tokens.T_STAR(!line,!line));
<INITIAL>"**"	=> (Tokens.T_STARSTAR(!line,!line));
<INITIAL>"-"	=> (Tokens.T_MINUS(!line,!line));
<INITIAL>"--"	=> (Tokens.T_MINUSMINUS(!line,!line));
<INITIAL>"?"	=> (Tokens.T_QUESTION(!line,!line));
<INITIAL>";;"	=> (Tokens.T_SEMISEMI(!line,!line));
<INITIAL>"@"	=> (Tokens.T_AT(!line,!line));
<INITIAL>"+"	=> (Tokens.T_PLUS(!line,!line));
<INITIAL>"++"	=> (Tokens.T_PLUSPLUS(!line,!line));
<INITIAL>"<+"	=> (Tokens.T_LESSPLUS(!line,!line));
<INITIAL>"+>"	=> (Tokens.T_PLUSGREATER(!line,!line));
<INITIAL>"|"    => (Tokens.T_VBAR(!line,!line));
<INITIAL>";"    => (Tokens.T_EOF(!line,!line));
<INITIAL>"."    => (Tokens.T_DOT(!line,!line));
<INITIAL>"!"    => (Tokens.T_BANG(!line,!line));
<INITIAL>","    => (Tokens.T_COMMA(!line,!line));
<INITIAL>"'"    => (Tokens.T_APOST(!line,!line));
<INITIAL>"="    => (Tokens.T_EQ(!line,!line));
<INITIAL>"=="   => (Tokens.T_DOUBLEEQ(!line,!line));
<INITIAL>"<="   => (Tokens.T_LEQ(!line,!line));
<INITIAL>"<"    => (Tokens.T_LEQ(!line,!line));
<INITIAL>"=>"   => (Tokens.T_DARROW(!line,!line));
<INITIAL>"/\\"	=> (Tokens.T_INTER(!line,!line));
<INITIAL>"("	=> (Tokens.T_LPAREN(!line,!line));
<INITIAL>")"	=> (Tokens.T_RPAREN(!line,!line));
<INITIAL>"["	=> (Tokens.T_LBRACK(!line,!line));
<INITIAL>"]"	=> (Tokens.T_RBRACK(!line,!line));
<INITIAL>"<"	=> (Tokens.T_LANGLE(!line,!line));
<INITIAL>">"	=> (Tokens.T_RANGLE(!line,!line));
<INITIAL>"{"	=> (Tokens.T_LCURLY(!line,!line));
<INITIAL>"}"	=> (Tokens.T_RCURLY(!line,!line));
<INITIAL>"->"	=> (Tokens.T_ARROW(!line,!line));
<INITIAL>"\\\\"	=> (Tokens.T_BIGLAMBDA(!line,!line));
<INITIAL>"\\"	=> (Tokens.T_LAMBDA(!line,!line));
<INITIAL>"\""   => (str_begin:=(!line); str_const:=[]; YYBEGIN STRING; lex());
<INITIAL>{id}   => (case yytext of
                       "use" => Tokens.T_USE(!line,!line)
		     | "type" => Tokens.T_TYPE(!line,!line)
		     | "set" => Tokens.T_SET(!line,!line)
		     | "reset" => Tokens.T_RESET(!line,!line)
		     | "debug" => Tokens.T_DEBUG(!line,!line)
		     | "check" => Tokens.T_CHECK(!line,!line)
		     | "T" => Tokens.T_NS(!line,!line)
		     | "NS" => Tokens.T_NS(!line,!line)
		     | "ns" => Tokens.T_NS(!line,!line)
		     | "with" => Tokens.T_WITH(!line,!line)
		     | "case" => Tokens.T_CASE(!line,!line)
		     | "bcase" => Tokens.T_BCASE(!line,!line)
		     | "of" => Tokens.T_OF(!line,!line)
		     | "as" => Tokens.T_AS(!line,!line)
		     | "for" => Tokens.T_FOR(!line,!line)
		     | "All" => Tokens.T_ALL(!line,!line)
		     | "Some" => Tokens.T_SOME(!line,!line)
		     | "open" => Tokens.T_OPEN(!line,!line)
		     | "pack" => Tokens.T_PACK(!line,!line)
		     | "hiding" => Tokens.T_HIDING(!line,!line)
		     | "begin" => Tokens.T_BEGIN (!line,!line)
		     | "end" => Tokens.T_END (!line,!line)
		     | "in" => Tokens.T_IN(!line,!line)
		     | "WIDTH" => Tokens.T_WIDTH(!line,!line)
		     | "STOP" => Tokens.T_STOP(!line,!line)
     | "FAIL" => Tokens.T_FAIL(!line,!line)
                     | _ => Tokens.T_ID (yytext,!line,!line));
<INITIAL>.	=> (error ("ignoring illegal character" ^ yytext,
			   !line,!line); lex());

<COMMENT>\n     => (next_line(); YYBEGIN INITIAL; lex());
<COMMENT>.	=> (lex());

<STRING>\n      => (next_line(); lex());
<STRING>"\""     => (YYBEGIN INITIAL;
		    Tokens.T_STR_CONST(implode(rev(!str_const)),
		    		       !str_begin,!line));
<STRING>.	=> (str_const:=(yytext::(!str_const)); lex());

