functor TEnv(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    sharing Typ.Globals = Globals
	    ) : TENV
	    = struct

structure Globals = Globals
structure Typ = Typ
open Globals
open Pp


datatype T = BND   of Id.T * Typ.T * T
	   | ABB   of Id.T * Typ.T * T
	   | VBND  of Id.T * Typ.T * T
	   | EMPTY 

(* ------------------------------------------------------------------------ *)

exception UnknownId of string

fun domsize (EMPTY) = 0
  | domsize (BND(_,_,te)) = 1 + (domsize te)
  | domsize (ABB(_,_,te)) = 1 + (domsize te)
  | domsize (VBND(_,_,te)) = 1 + (domsize te)

exception TriedToPopEmptyTEnv
fun pop (EMPTY) = raise TriedToPopEmptyTEnv
  | pop (BND(_,_,te)) = te
  | pop (ABB(_,_,te)) = te
  | pop (VBND(_,_,te)) = te

fun popn te 0 = te
  | popn te n = popn (pop te) (n-1) 

fun index bvs i =
  let fun ind (EMPTY) n =
            raise UnknownId(Id.tostr i)
        | ind (BND(i',_,rest)) n =
            if Id.== i i'
               then n
               else ind rest (n+1)
        | ind (ABB(i',_,rest)) n =
            if Id.== i i'
               then n
               else ind rest (n+1)
        | ind (VBND(i',_,rest)) n =
            if Id.== i i'
               then n
               else ind rest (n+1)
  in ind bvs 0
  end

exception TypeVariableOutOfRange of int

fun lookup_name te i = 
  let fun l (EMPTY) i = Id.intern(("<BAD INDEX: " ^ (makestring i) ^ ">"))
        | l (BND(name,_,rest)) i =
              if i = 0 then name else (l rest (i-1))
        | l (ABB(name,_,rest)) i =
              if i = 0 then name else (l rest (i-1))
        | l (VBND(name,_,rest)) i =
              if i = 0 then name else (l rest (i-1))
  in l te i
  end

fun bare_lookup te n fbnd fvbnd fabb =
  let fun l (EMPTY) i = 
              raise TypeVariableOutOfRange(n)
        | l (BND(_,t,rest)) i =
              if i = 0 
                then fbnd t
                else (l rest (i-1))
        | l (VBND(_,t,rest)) i =
              if i = 0 
                then fvbnd t
                else (l rest (i-1))
        | l (ABB(_,t,rest)) i =
              if i = 0 
                then fabb t
                else (l rest (i-1))
  in l te n
  end

fun lookup_and_relocate te n fbnd fvbnd fabb =
  let fun l (EMPTY) i = 
              raise TypeVariableOutOfRange(n)
        | l (BND(_,t,rest)) i =
              if i = 0 
                then fbnd (Typ.relocate (n+1) t)
                else (l rest (i-1))
        | l (VBND(_,t,rest)) i =
              if i = 0 
                then fvbnd (Typ.relocate (n+1) t)
                else (l rest (i-1))
        | l (ABB(_,t,rest)) i =
              if i = 0 
                then fabb (Typ.relocate (n+1) t)
                else (l rest (i-1))
  in l te n
  end

(* ----------------------------------------------------------------------- *)

fun push_bound te i t = BND(i,t,te)
fun push_abbrev te i t = ABB(i,t,te)
fun push_binding te i t = VBND(i,t,te)
val empty = EMPTY

(* ----------------------------------------------------------------------- *)

exception WrongKindOfId of T * int * string

fun lookup_and_relocate_bound te i = 
  lookup_and_relocate te i 
    (fn b => b)
    (fn _ => raise WrongKindOfId(te,i,"tvar"))
    (fn _ => raise WrongKindOfId(te,i,"tvar"))

fun lookup_and_relocate_binding te i = 
  lookup_and_relocate te i 
    (fn _ => raise WrongKindOfId(te,i,"var"))
    (fn b => b)
    (fn _ => raise WrongKindOfId(te,i,"var"))

end

