functor TopLevel(
	 structure Globals: GLOBALS
	 structure TEnv: TENV
	 structure TypHelper: TYPHELPER
	 structure TrmHelper: TRMHELPER
	 structure ParseRes: PARSERES
	 structure Leq: LEQ
	 structure Synth: SYNTH
	 structure Abbrev: ABBREV
	 structure Simplify: SIMPLIFY
	 sharing TypHelper.Globals = Globals
	     and ParseRes.TypHelper = TypHelper
	     and ParseRes.TrmHelper = TrmHelper
	     and Leq.TEnv = TEnv
	     and Abbrev.TEnv = TEnv
	     and Simplify.TEnv = TEnv
	     and Synth.TEnv = TEnv
	     and TypHelper.TEnv = TEnv
	     and Synth.Leq = Leq
	     and Synth.Trm = TrmHelper.Trm
	     and Simplify.Typ = TypHelper.Typ
	) : TOPLEVEL = struct

structure ParseRes = ParseRes

open Globals;
open ParseRes;

val ie = createflag ["ie", "origtypes"] false
val recap = createflag ["recap"] false
val origterms = createflag ["origterms"] false

type TopLevelData = TEnv.T 

fun init () = TEnv.empty

fun do_command (tld as te) (continue: TopLevelData->'a) (error: TopLevelData->'a) cmd = 
    (case cmd of
         Type_Assumption(i,pt) =>
       	   let val ot = TypHelper.debruijnify te pt
       	       val t = Abbrev.collapse te (Simplify.squash te ot)
           in 
              if !recap 
                then (
		   write (Id.tostr i);
		   write " < ";
		   TypHelper.prt_outer te t;
		   write "\n")
	         else ();
       	      continue (TEnv.push_bound te i t)
           end
       | Type_Abbrev(i,pt) =>
       	   let val ot = TypHelper.debruijnify te pt
       	       val t = Abbrev.collapse te (Simplify.squash te ot)
           in 
              if !recap 
                then (
		  write (Id.tostr i);
		  write " == ";
		  TypHelper.prt_outer te t;
		  write "\n" 
		  (*;
		  write " == ";
		  TypHelper.prt_nested te t;
		  write "\n" *))
	        else ();
	      continue (TEnv.push_abbrev te i t)
           end
       | Term_Def(i,ptrm) =>
       	   let val trm = TrmHelper.debruijnify te ptrm
       	       val _ = Synth.setquerytenv te
       	       val otyp = Synth.synth te trm
       	       val styp = Simplify.squash te otyp
       	       val typ = Abbrev.collapse te styp
           in 
              write (Id.tostr i);
	      if !origterms
	         then (
		    write " = ";
		    setb ();
		    TrmHelper.prt te trm;
		    break true ~3;
		    endb()) 
	         else ();
              write " : ";
              setb();
	      if !ie 
                then (TypHelper.prt_outer te otyp;
                      write "  ";
                      ubreak ~5;
                      write "i.e. ")
                else ();
              TypHelper.prt_outer te typ;
              endb();
              write "\n";
              continue (TEnv.push_binding te i typ)
           end
       | Term_Assumption(i,ptp) =>
       	   let val ot = TypHelper.debruijnify te ptp
       	       val typ = Abbrev.collapse te (Simplify.squash te ot)
           in 
              if !recap
                then (
		  write (Id.tostr i);
		  write " : ";
		  TypHelper.prt_outer te typ;
		  write "\n")
		else ();
              continue (TEnv.push_binding te i typ)
           end
       | Leq(pt1,pt2) =>
           let val t1 = TypHelper.debruijnify te pt1
               val t2 = TypHelper.debruijnify te pt2
           in 
              if Leq.leq te t1 t2 
                then write "Yes\n"
                else write "No\n";
	      continue tld
           end
    )
    handle 
      TEnv.WrongKindOfId(te,i,which) => 
      	  (write ("Wrong kind of identifier: "^ (makestring i) ^" ("
      	  	  ^ which ^ " expected)\nin "); 
      	   TypHelper.prt_tenv te;
       	   error tld)
    | TEnv.UnknownId(s) =>
          (write ("Unbound id: " ^ s ^ "\n");
           error tld)

end


