functor Typ(
	    structure Globals: GLOBALS
	    ) : TYP
	    = struct

structure Globals = Globals
open Globals
open Pp

(* $$$ Just noticed some problems here:

     - do_relocs is too eager: this leads to a lot of inefficiency
     - eqq should take abbreviations into account
     - occurs_free_in is wrong
*)


(* ----------------------------------------------------------------------- *)

(* A list of (CUTOFF,OFFSET) pairs, in innermost-first order *)
type relocs = (int * int) list

datatype T = 
	    TVAR of int
	  | ARROW of T * T
	  | ALL of Id.T * T * T
	  | SOM of Id.T * T * T
	  | MEET of T list
	  | OBJ of Id.T * T 
	  | MINUS of T * (Id.T list)
	  | RELOC of relocs * T

(* Play with optimizing these in simple ways (as for reloc) *)
fun tvar i = TVAR(i)
fun arrow t1 t2 = ARROW(t1,t2)
fun all n t1 t2 = ALL(n,t1,t2)
fun some n t1 t2 = SOM(n,t1,t2)
fun meet ts = MEET(ts)
fun obj n t = OBJ(n,t)
fun minus t ids = MINUS(t,ids)
fun reloc [(c,0)] t = t
  | reloc [(0,off)] (RELOC([(0,oldoff)],t)) = reloc [(0,off+oldoff)] t
  | reloc r t = RELOC(r,t)


(* ------------------------------------------------------------------------ *)

fun do_relocs_var [] (i:int) = 
	i
  | do_relocs_var ((cut,off)::tl) i = 
  	do_relocs_var tl (if i>=cut then i+off else i)

fun inc_cuts (r : relocs) = 
    map (fn (cut,off) => (cut+1,off)) r

fun append_relocs (r1 as [(cut1,off1)]) (r2 as ((cut2,off2)::r2tl)) = 
    if cut1 = cut2
      then if off1+off2 = 0
             then r2tl
             else (cut1,off1+off2)::r2tl
      else r1 @ r2
  | append_relocs r1 r2 = r1 @ r2

fun do_relocs t = let
  fun dr r (TVAR (i)) = tvar (do_relocs_var r i)
    | dr r (ARROW(t1,t2)) = arrow (dr r t1) (dr r t2)
    | dr r (ALL(n,t1,t2)) = all n (dr r t1) (dr (inc_cuts r) t2)
    | dr r (SOM(n,t1,t2)) = some n (dr r t1) (dr (inc_cuts r) t2)
    | dr r (MEET(ts)) = meet (map (fn t => dr r t) ts)
    | dr r (OBJ(n,t)) = obj n (dr r t)
    | dr r (MINUS(t,ids)) = minus (dr r t) ids
    | dr r (RELOC(rt,t)) = dr (append_relocs rt r) t
  in dr [] t
  end

fun relocate_above cut 0   t = t
  | relocate_above cut off t = reloc [(cut,off)] t

fun relocate off t = relocate_above 0 off t

fun relocate_now off t = do_relocs (relocate_above 0 off t)

datatype Exploded = 
	    EXPTVAR of int
	  | EXPARROW of T * T
	  | EXPALL of Id.T * T * T
	  | EXPSOME of Id.T * T * T
	  | EXPMEET of T list
	  | EXPOBJ of Id.T * T 
	  | EXPMINUS of T * (Id.T list)

fun explode t = let
  fun dig r (RELOC(rt,t)) = dig (append_relocs rt r) t
    | dig r t = (r,t)
  val (r,t') = dig [] t
  in if (null r)
    then case t' of
	   TVAR(i) => EXPTVAR (i)
	 | ARROW(t1,t2) => EXPARROW (t1,t2)
	 | ALL(n,t1,t2) => EXPALL (n,t1,t2)
	 | SOM(n,t1,t2) => EXPSOME (n,t1,t2)
	 | MEET(ts) => EXPMEET (ts)
	 | OBJ(n,t) => EXPOBJ (n,t)
	 | MINUS(t,ids) => EXPMINUS (t,ids)
	 | _ => raise CantHappen
    else case t' of
	   TVAR(i) => EXPTVAR (do_relocs_var r i)
	 | ARROW(t1,t2) => EXPARROW (reloc r t1, reloc r t2)
	 | ALL(n,t1,t2) => EXPALL (n, reloc r t1, reloc (inc_cuts r) t2)
	 | SOM(n,t1,t2) => EXPSOME (n, reloc r t1, reloc (inc_cuts r) t2)
	 | MEET(ts) => EXPMEET (map (fn t => reloc r t) ts)
	 | OBJ(n,t) => EXPOBJ (n, (reloc r t))
	 | MINUS(t,ids) => EXPMINUS ((reloc r t), ids)
	 | _ => raise CantHappen
  end


(* ---------------------------------------------------------------------- *)
(* $$$ Can cut down on consing here by folding in the do_relocs *)

fun tsubst_top targ tbody = let
  fun s i (t as TVAR(i')) = 
	if i = i' 
	  then relocate_above 0 i targ
	else if i < i'
	  then tvar (i'-1)
	else t
    | s i (ARROW(t1,t2)) = arrow (s i t1) (s i t2)
    | s i (ALL(x,t1,t2)) = all x (s i t1) (s (i+1) t2)
    | s i (SOM(x,t1,t2)) = some x (s i t1) (s (i+1) t2)
    | s i (MEET(ts)) = meet (map (fn t => s i t) ts)
    | s i (OBJ(n,t)) = obj n (s i t) 
    | s i (MINUS(t,ids)) = minus (s i t) ids
    | s i (RELOC _) = raise CantHappen
  in s 0 (do_relocs tbody)
  end

fun occurs_free_in v t = let
  fun s i (TVAR(i')) = (v = i')
    | s i (ARROW(t1,t2)) = (s i t1) orelse (s i t2)
    | s i (ALL(x,t1,t2)) = (s i t1) orelse (s (i+1) t2)
    | s i (SOM(x,t1,t2)) = (s i t1) orelse (s (i+1) t2)
    | s i (MEET(ts)) = forsome (fn t => s i t) ts
    | s i (OBJ(n,t)) = (s i t) 
    | s i (MINUS(t,ids)) = (s i t) 
    | s i (RELOC _) = raise CantHappen
  in s v (do_relocs t)
  end

fun eqq (TVAR(i)) (TVAR(j)) = (i = j)
  | eqq (ARROW(s1,s2)) (ARROW(t1,t2)) = (eqq s1 t1) andalso (eqq s2 t2)
  | eqq (ALL(_,s1,s2)) (ALL(_,t1,t2)) = (eqq s1 t1) andalso (eqq s2 t2)
  | eqq (SOM(_,s1,s2)) (SOM(_,t1,t2)) = (eqq s1 t1) andalso (eqq s2 t2)
  | eqq (MEET(ss)) (MEET(ts)) = 
  	  ((length ss) = (length ts))
          andalso (forall (fn (s,t) => eqq s t) (zip ss ts))	
  | eqq (OBJ(ns,s)) (OBJ(nt,t)) = (Id.== ns nt) andalso (eqq s t)
  | eqq (MINUS(s,idss)) (MINUS(t,idst)) = 
          (eqq s t) andalso forallpairs (fn (is,it) => Id.== is it) idss idst
  | eqq _ _ = false

fun eq t1 t2 = eqq (do_relocs t1) (do_relocs t2)

(* ------------------------------------------------------------------------- *)

val NS = meet []

fun prt_exactly t = let
      fun prt_relocs [] = write "nil"
        | prt_relocs [(cut,off)] = 
            (write (makestring cut);
             write "/";
             write (makestring off))
        | prt_relocs ((cut,off)::tl) = 
            (write (makestring cut);
             write "/";
             write (makestring off);
             write ",";
             prt_relocs tl)
      fun p (TVAR(i)) =
             (write (makestring i))
        | p (ARROW(t1,t2)) =
              (write "(";
               p t1;
               write "->";
               p t2;
               write ")")
        | p (ALL(i,t1,t2)) =
              (write "(All ";
               write (Id.tostr i);
               write "<";
               p t1;
               write ". ";
               p t2;
               write ")")
        | p (SOM(i,t1,t2)) =
              (write "(Some ";
               write (Id.tostr i);
               write "<";
               p t1;
               write ". ";
               p t2;
               write ")")
        | p (MEET([])) =
               write "NS"
        | p (MEET(ts)) =
              (write "/\\[";
               plist ts;
               write "]")
        | p (OBJ(n,t)) =
              (write (Id.tostr n);
               write ":";
               p t)
        | p (MINUS(t,ids)) =
              (p t;
               write "\\{";
               mapunit_tuple 
               	    (fn i => write (Id.tostr i)) (fn()=> write",")
               	    ids;
	       write "}")
	| p (RELOC(r,t)) =
	      (write "{";
	       prt_relocs r;
	       write ":";
	       p t;
	       write "}")
      and plist [] = 
            ()
        | plist [t] = 
            p t
        | plist (hd::tl) = 
            (p hd; write ","; plist tl)
  in p t
  end

end

