% %%%%%%%%%%%%%%%%%%%%
% %%%    Basics    %%%
% %%%%%%%%%%%%%%%%%%%%

% Some useful abbreviations for maximal types
Top1 = Top(*);
Top2 = Top(*->*);
Top3 = Top(*->*->*);
top = <<"()" : Top1>>;
printer <<"fn x:int => fn() => (write (\"<val>\"))"
        : Top1->Top1>>;

% The one-point type (and its one point)
Unit < Top1;
unit = <<"()" : Unit>>;

% Booleans
Bool < Top1;
primif = <<"fn b => fn()=> fn e1 => fn()=> fn e2 => 
              if b() then e1 else e2"
         : All(X<Top1) Bool -> X -> X -> X >>;
true = <<"true" : Bool>>;
false = <<"false" : Bool>>;
not = fun(b:Bool) primif Bool b false true;
and = fun(b1:Bool) fun(b2:Bool) primif Bool b1 b2 false;
or = fun(b1:Bool) fun(b2:Bool) primif Bool b1 true b2;
printer <<"fn b => fn() => if b() then write \"true\" else write \"false\"" 
        : Bool->Top1>>;

% Strings
String < Top1;
write = <<"fn s:unit->string => fn() => write (s())": String -> Unit>>;
printer <<"fn x => fn() => write (\"\\"\" ^ (x()) ^ \"\\"\")" : String->Top1>>;
concat = <<"fn s => fn()=> fn t => fn()=> (s()) ^ (t())" 
         : String->String->String >>;

% Integers
Int < Top1;
printer <<"fn x:unit->int => fn() => (write (makestring (x())))"
         : Int->Top1>>;
int_to_string = <<"fn x:unit->int => fn()=> makestring (x())" 
                : Int -> String>>;
plus = <<"(fn x:unit->int => fn()=> fn y:unit->int => 
             fn()=> (x()) + (y()))" 
         : Int->Int->Int >>;
minus = <<"(fn x:unit->int => fn()=> fn y:unit->int => 
               fn()=> (x()) - (y()))" 
         : Int->Int->Int >>;
times = <<"(fn x:unit->int => fn()=> fn y:unit->int => 
              fn()=> (x()) * (y()))" 
         : Int->Int->Int >>;
div = <<"(fn x:unit->int => fn()=> fn y:unit->int => 
            fn()=> (x()) div (y()))" 
         : Int->Int->Int >>;
succ = fun(i:Int) plus i 1;
pred = fun(i:Int) plus i ~1;
% Should probably nuke this one:
eq = <<"(fn x:unit->int =>  fn()=> fn y:unit->int => 
           fn()=> (x()) = (y()))" 
       : Int->Int->Bool >>;
eqInt = fun(x:Int) fun(y:Int) eq x y;

% Recursion
%   y : (All(A) #((A->A)->A))
%     = All(A) unit -> (unit->(unit->(unit->A)->(unit->A))) -> (unit->A)
rec = <<"let fun y () f () = f () (y () f) () in y () end" 
      : All(A<Top(*)) (A->A) -> A >>;

% Other basic data types
Color < Top1;
color_to_string = <<"fn x => x" : Color -> String>>;
red = <<"\"red\"" : Color>>;
yellow = <<"\"yellow\"" : Color>>;
blue = <<"\"blue\"" : Color>>;
black = <<"\"black\"" : Color>>;
eqColor = <<"(fn x:unit->string =>  fn()=> fn y:unit->string => 
                fn()=> (x()) = (y()))" 
          : Color->Color->Bool >>;
printer <<"fn x => fn() => write (x())"
         : Color->Top1>>;

% Lists
List = Fun(E) All(A) (E->A->A) -> A -> A;
nil = fun(E) fun(A) fun(f:E->A->A) fun(a:A) a;
cons = fun(E) fun(e:E) fun(l:List(E)) 
         fun(A) fun(f:E->A->A) fun(a:A) 
           f e (l A f a);
iterList = fun(E) fun(l:List(E)) l;


