functor ParseRes 
	    (structure TypHelper: TYPHELPER
	     structure TrmHelper: TRMHELPER
	     structure Globals: GLOBALS
	     sharing TypHelper.Globals = Globals
		 and TypHelper.Id = Globals.Id
		 and TrmHelper.Id = Globals.Id
		 and TrmHelper.TypHelper = TypHelper
	         and TrmHelper.Trm.TypHelper = TypHelper
	    ) : PARSERES 
	    = struct

structure TypHelper = TypHelper
structure TrmHelper = TrmHelper
structure Globals = Globals

open TrmHelper
open TypHelper
open Globals

datatype SystemSpecific = 
    Leq of TypHelper.PreTyp * TypHelper.PreTyp
  | Eqv of TypHelper.PreTyp * TypHelper.PreTyp
  | Type_Assumption of Globals.Id.T * TypHelper.PreTyp
  | Type_Abbrev of Globals.Id.T * TypHelper.PreTyp
  | Term_Def of Globals.Id.T * TrmHelper.PreTrm
  | Term_Assumption of Globals.Id.T * TypHelper.PreTyp
  | PrinterDecl of TrmHelper.PreTrm
  | ObjectTypeDecl of 
       Globals.Id.T 
     * Globals.Id.T 
     * ((Globals.Id.T * TypHelper.PreTyp) list)
  | ClassDecl of 
       Globals.Id.T 
     * Globals.Id.T 
     * Globals.Id.T 
     * Globals.Id.T 
     * (Globals.Id.T * TypHelper.PreTyp * TrmHelper.PreTrm) list
     * Globals.Id.T list
     * Globals.Id.T list
     * (Globals.Id.T * TrmHelper.PreTrm) list

datatype Item =
    Specific of SystemSpecific
  | Use of string
  | Set of string * string
  | Cmd of string * (string option)
  | Nothing

datatype T = ITEMS of Item list

(* ---------------------------------------------------------------------- *)

val M = Id.intern "M"
val Object = PRETVAR(Id.intern "Object")
val ObjectOfM = PREAPPT(Object, PRETVAR(M))
val oname = Id.intern "o"
val methods = Id.intern "methods"
val state = Id.intern "state"
val x = Id.intern "x"

datatype Actions = IGNORE
                 | WARN
                 | WARNREPEATEDLY
                 | FAIL

val oldSyntaxAction = ref(WARNREPEATEDLY)

fun change_action (SOME("ignore")) = oldSyntaxAction := IGNORE
  | change_action (SOME("warn")) = oldSyntaxAction := WARN
  | change_action (SOME("warnrepeatedly")) = oldSyntaxAction := WARNREPEATEDLY
  | change_action (SOME("fail")) = oldSyntaxAction := FAIL
  | change_action (_) = oldSyntaxAction := IGNORE

val _ = register_cmd "oldsyntax" change_action;

exception OldSyntax

fun warn(s) = write ("*** Warning: Old syntax: "^s^" ***\n");

fun oldSyntax s = 
  case !oldSyntaxAction of
      IGNORE => ()
    | WARN => (warn(s); oldSyntaxAction := IGNORE)
    | WARNREPEATEDLY => warn(s)
    | FAIL => (warn(s ^ "...Failing!"); raise OldSyntax)
      (* This one doesn't seem to work!  Probably the parser is catching
         the error instead of letting it bubble out to the top-level. *)

end 
