functor Trm(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    structure TypHelper: TYPHELPER
	    sharing Typ.Globals = Globals
	        and TypHelper.Typ = Typ
	        and TypHelper.TEnv = TEnv
	        and TEnv.Typ = Typ
	        and Globals.Id = Typ.Id
	        and Typ.Id = TEnv.Id
	        and TEnv.Id = TypHelper.Id
	    ) : TRM
	    = struct

structure Globals = Globals
structure Typ = Typ
structure TypHelper = TypHelper
structure TEnv = TEnv
open Globals

type Prim = string

datatype T = 
	    VAR of int
	  | ABS of Id.T * Typ.T * T
	  | APP of T * T
	  | TABS of Id.T * Typ.T * T
	  | TAPP of T * Typ.T
          | RCDV of (Id.T * T) list
	  | PROJ of T * Id.T 
	  | OPEN of Id.T * Id.T * T * T
	  | PACK of T * Typ.T * Typ.T 
	  | ASSERT of T * Typ.T
	  | ABBREV of Id.T * Typ.T * T
	  | LET of Id.T * T * T
	  | PRIM of Prim * Typ.T

fun relocate_above cut off e = let
      fun ra c (VAR(i))        = VAR (if i>=c then i+off else i)
        | ra c (ABS(n,t,e))    = ABS (n, (Typ.relocate_above c off t),
                                         (ra (c+1) e))
        | ra c (APP(e1,e2))    = APP (ra c e1, ra c e2)
        | ra c (TABS(n,t,e))   = TABS (n, (Typ.relocate_above c off t),
                                         (ra (c+1) e))
        | ra c (TAPP(e1,t2))   = TAPP (ra c e1, Typ.relocate_above c off t2)
        | ra c (RCDV(fields))  = RCDV (map (fn(l,e)=> (l, ra c e)) fields)
        | ra c (PROJ(e,l))     = PROJ (ra c e, l)
        | ra c (OPEN(a,x,e,b)) = OPEN (a, x, ra c e, ra (c+2) b)
        | ra c (PACK(e,t,h))   = PACK (ra c e, 
                                       Typ.relocate_above c off t,
                                       Typ.relocate_above c off h)
        | ra c (ASSERT(e,t))   = ASSERT (ra c e, Typ.relocate_above c off t)
        | ra c (ABBREV(a,t,e)) = ABBREV (a, 
                                         Typ.relocate_above c off t,
                                         ra (c+1) e)
        | ra c (LET(x,e,b))    = LET (x, ra c e, ra (c+1) b)
        | ra c (PRIM(p,t))     = PRIM (p, Typ.relocate_above c off t)
      in ra cut e
      end

fun relocate off t = relocate_above 0 off t

fun subst_top arg body = let
  fun bump c t = Typ.relocate_above 1 ~1 t
  fun ts i (VAR(i'))          = if i=i' 
   				  then (relocate i arg) 
  				else if i < i'
  				  then VAR(i'-1)
  				else VAR(i')
    | ts i (ABS(n,t,e))       = ABS (n, bump i t, (ts (i+1) e))
    | ts i (APP(e1,e2))       = APP (ts i e1, ts i e2)
    | ts i (TABS(n,t,e))      = TABS (n, bump i t, (ts (i+1) e))
    | ts i (TAPP(e1,t2))      = TAPP (ts i e1, bump i t2)
    | ts i (RCDV(fields))     = RCDV (map (fn (l,e) => (l, ts i e)) fields)
    | ts i (PROJ(e,l))        = PROJ (ts i e, l)
    | ts i (OPEN(a,x,e,b))    = OPEN (a, x, ts i e, ts (i+2) b)
    | ts i (PACK(e,t,h))      = PACK (ts i e, bump i t, bump i h)
    | ts i (ASSERT(e,t))      = ASSERT (ts i e, bump i t)
    | ts i (ABBREV(a,t,e))    = ABBREV (a, bump i t, ts (i+1) e)
    | ts i (LET(x,e,b))       = LET (x, ts i e, ts (i+1) b)
    | ts i (PRIM(p,t))        = PRIM (p, bump i t)
  in ts 0 body
  end

fun subst_top_type typ body = let
  fun ts i (VAR(i'))          = if i < i'
  				  then VAR(i'-1)
  				  else VAR(i')
    | ts i (ABS(n,t,e))       = ABS (n, Typ.tsubst typ i t, (ts (i+1) e))
    | ts i (APP(e1,e2))       = APP (ts i e1, ts i e2)
    | ts i (TABS(n,t,e))      = TABS (n, Typ.tsubst typ i t, (ts (i+1) e))
    | ts i (TAPP(e1,t2))      = TAPP (ts i e1, Typ.tsubst typ i t2)
    | ts i (RCDV(fields))     = RCDV (map (fn (l,e) => (l, ts i e)) fields)
    | ts i (PROJ(e,l))        = PROJ (ts i e, l)
    | ts i (OPEN(a,x,e,b))    = OPEN (a, x, ts i e, ts (i+2) b)
    | ts i (PACK(e,t,h))      = PACK (ts i e, 
                                      Typ.tsubst typ i t, 
                                      Typ.tsubst typ i h)
    | ts i (ASSERT(e,t))      = ASSERT (ts i e, Typ.tsubst typ i t)
    | ts i (ABBREV(a,t,e))    = ABBREV (a, Typ.tsubst typ i t, ts (i+1) e)
    | ts i (LET(x,e,b))       = LET (x, ts i e, ts (i+1) b)
    | ts i (PRIM(p,t))        = PRIM (p, Typ.tsubst typ i t)
  in ts 0 body
  end


fun prt_primitive s = write ("\"" ^ s ^ "\"")

fun prt_exactly e = let
  fun pe (VAR(i)) = 
      write (makestring i)
    | pe (ABS(n,t,e)) = 
      (write "(fun(";
       write (Id.tostr n);
       write ":";
       Typ.prt_exactly t;
       write ") ";
       prt_exactly e;
       write ")")
    | pe (APP(e1,e2)) = 
      (write "(";
       pe e1;
       write " ";
       pe e2;
       write ")")
    | pe (TABS(n,t,e)) = 
      (write "(fun(";
       write (Id.tostr n);
       write "<";
       Typ.prt_exactly t;
       write ") ";
       prt_exactly e;
       write ")")
    | pe (TAPP(e1,t2)) = 
      (write "(";
       pe e1;
       write " [";
       Typ.prt_exactly t2;
       write "])")
    | pe (RCDV(fields)) =
      (write "{";
       mapunit_tuple (fn (l,e) => (write (Id.tostr l);
                                   write "=";
                                   pe e))
                     (fn () => write ", ")
                     fields;
       write "}")
    | pe (PROJ(e,l)) =
      (pe e;
       write ".";
       write (Id.tostr l))
    | pe (OPEN(a,x,e,b)) =
      (write "open ";
       pe e;
       write " as [";
       write (Id.tostr a);
       write ",";
       write (Id.tostr x);
       write "] in ";
       prt_exactly b;
       write " end")
    | pe (PACK(e,t,h)) =
      (write "pack ";
       pe e;
       write " as ";
       Typ.prt_exactly t;
       write " hiding ";
       Typ.prt_exactly h;
       write " end")
    | pe (ASSERT(e,t)) =
      (write "assert ";
       pe e;
       write " : ...")
    | pe (ABBREV(a,t,e)) =
      (write "let ";
       write (Id.tostr a);
       write " == ";
       Typ.prt_exactly t;
       write " in ";
       prt_exactly e;
       write " end")
    | pe (LET(x,e,b)) =
      (write "let ";
       write (Id.tostr x);
       write " = ";
       pe e;
       write " in ";
       prt_exactly b;
       write " end")
    | pe (PRIM(p,t)) =
      (write "<<";
       prt_primitive p;
       write " : ";
       Typ.prt_exactly t;
       write ">>")
  in pe e
  end

end
